; dervish: unflappable, ascetic, source of wisdom.
; this thing has all of the above characteristics, and
; whirls madly between decoding infrared remote control signals,
; handling a hd44780 lcd and doing bidirectional serial comms.

; copyright (c) 2007-2011 Alexander Zangerl <az@snafu.priv.at>
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation (version 2 of the License).
;
; $Id: dervish.asm,v 2.5 2011/04/25 04:17:49 az Exp $

processor p16f628a
#include "p16f628a.inc"
#include "highlevel.inc"
	RADIX hex

;; carry 0 means negative, carry 1 zero or positive
#define JNEG	BNC
#define JNNEG	BC	

; lcd control pins/bits
; rb4..7 are lcd data.
RSPIN	equ	1		; ra1, top pin
BLPIN	equ	0		; ra0, second pin
ENAPIN	equ	7		; ra7, third pin

; a piezo beeper for doing an alarm clock
SQPIN	equ	6		; ra6, fourth pin

	__CONFIG  _INTRC_OSC_NOCLKOUT & _BODEN_OFF & _CP_OFF & _MCLRE_ON & _WDT_OFF & _LVP_OFF

	cblock 0x20
	scratch			; tempvar for delay routines
	
	scritch			; other tempvars
	foo
	baz

	rbe			; temp var for fifo code

	; inthandler variables
	bar			; infrared inthandler
	togo			; ditto, bits to go
	d1			; remote code bytes, filled in by inthandler (when irdone unset)
	d2			; note: d1..4 must be consecutive
	d3			
	d4

	; set by inthandler, reset by worker loop. bit definitions below
	irstatus		; idle/haveresult/haverepeat
	
	ircode1			; saved bytes 3 and four and counter
	ircode2
	ircount
	
	cmdstatus		; serial command states

	rcvr:9,rcvw		; serial receive fifo, 8 chars deep
	txr:.17,txw		; serial transmit fifo, 16 chars deep
	endc

	cblock	0x70		; bank-shared vars
	w_save			; interrupt context
	fsr_save
	status_save
	endc

; bits in irstatus	
IRIDLE	equ	7		; set when ready for hdr/rpt, cleared when reading data bits
IRDONE	equ	6		; set when d1...d4 ready for collection, cleared when collected
IRRPT	equ	5		; set when repeat code received, cleared when collected
IRERR	equ	4		; set when inthandler aborts an ir receive


; cmdstatus: remembers how to handle next serial char
CMDBIT	equ	0		; next char is cmd
BACKBIT	equ	1		; next char is backlight command
BEEPBIT equ	2		; next char is beeper command
DATABIT	equ	3		; next char is data, no ifs and buts

; these were chosen as not usually useful as data bytes: they are not printable
; however, to be absolutely safe, there is a data escape, too.
CMDESC	equ	0x80		; ser cmd esc:	 next char is command
BACKESC	equ	0x81		; next char is backlight state
BEEPESC	equ	0x82		; next char is beeper state
DATAESC	equ	0x83		; next char is definitely data

SENDNIB	macro value
	movlw	value
	call	sendnibble
	endm

; carry is used as is-command
SEND	macro value,iscmd
	if (iscmd)
		bsf	STATUS,C
	else
		bcf	STATUS,C
	endif
	movlw	value
	call	sendlcd
        if (iscmd && value<=0x03) ; clear, home take longer!
	movlw	2
        call    delayXms
        endif
	endm

	org 0			; power-on or reset
	goto main
	org 0x4			; int vector
	movwf	w_save		; save context
	swapf	STATUS,W
	bcf	STATUS,RP0	; bank0
	movwf	status_save
	movfw	FSR
	movwf	fsr_save
	
	call	inthandler

	movfw	fsr_save	
	movwf	FSR
	swapf	status_save,W
	movwf	STATUS
	swapf	w_save,F
	swapf	w_save,W
	retfie

; expects value in w, uart on, bank 0
; leaves w as is.
; note:	this busywaits until the uart is ready!
sendser:
txtest:	btfss	PIR1,TXIF	; ready to transmit? txif is 1 if free
	goto	txtest
	movwf	TXREG		; ser out
	nop
	return

#include "fifo.inc"	

; startup chores: setup io and peripherals, clean memory
main:				
	movlw	7
	movwf	CMCON		; comparators are on by default!
	bsf	T1CON,TMR1ON	; disable timer1
	clrf	TMR0		; blank timer0 and prescaler

	; setup fifos
	FINIT	rcvr,rcvw
	FINIT	txr,txw
	
	bsf	STATUS,RP0	; bank1
	bcf	OPTION_REG,INTEDG ; need rb0 int on falling edge
	bcf	OPTION_REG,PS0	; 110: 1/128 prescale
	bcf	OPTION_REG,PSA	; prescaler to tmr0
	bcf	OPTION_REG,T0CS	; timer0 now on with internal clock
	
	; rs, backlight, ena, beeper on ra are outputs
	movlw	~((1<<RSPIN)|(1<<ENAPIN)|(1<<BLPIN)|(1<<SQPIN)) 
	movwf	TRISA
	movlw	0x0f ; upper half of rb cleared for output
	movwf	TRISB

	; usart: 19200, 38400 are doable without ext. osc
	; 1920cps is ample i think, so we go for 19200.
	bsf	TXSTA,TXEN	; transmitter on
	bcf	TXSTA,SYNC	; make sure async
	movlw	.12		; 19200
	bsf	TXSTA,BRGH	; hispeed
	movwf	SPBRG		;
	bsf	PIE1,RCIE	; receive interrupt on
	
	bcf	STATUS,RP0	; bank0

	clrf	PORTA		; setup output options first
	clrf	PORTB

	bsf	PORTA,SQPIN	; squeek until setup done

	; (re)init the lcd, as per datasheet procedure
	call	delay100ms	; wait 100ms
	SENDNIB	0x30
	movlw	5		; min. 4.1ms as per datasheet
	call	delayXms
	SENDNIB 0x30
	movlw	.25
	call	delayX5us	; min. 100us
	SENDNIB 0x30
	SENDNIB 0x20		; done!
	
	SEND	0x28,1		; 2 bit, 2 line mode
	SEND	0x08,1		; off
	SEND	0x01,1		; clear
	SEND	0x06,1		; on

	SEND	0x1,1		; clear display
	; lcd init done.

	
	; get the ir stuff ready; we should not get any spurious interrupts
	; after waiting for the lcd to start
	clrf	irstatus	; nothing read yet, d1-4 invalid
	bsf	irstatus,IRIDLE	; but we're ready
	clrf	ircode1		
	clrf	ircode2
	clrf	ircount
	
	bcf	INTCON,T0IF	; clear tmr0 overflow int flag
	bcf	INTCON,INTF	; clear rb0 int flag

	bsf	INTCON,INTE	; interrupt on rb0 on
	bsf	INTCON,PEIE	; peripheral ints on for serial
	bsf	INTCON,GIE	; global ints on

	clrf	cmdstatus	; nothing received yet
	bsf	RCSTA,CREN	; enable serial receiver (tx is already on before lcd)
	bsf	RCSTA,SPEN 	; and enable the serial port overall

	; startup msg output to serial
	movlw	startmsg
	movwf	foo		; save in foo
msg_cont:
	movfw	foo
	bsf	STATUS,RP0
	movwf	EEADR
	bsf	EECON1,RD
	movfw	EEDATA
	bcf	STATUS,RP0
	andlw	0xff		; \0, end of message?
	btfsc	STATUS,Z	; skip if nonzero
	goto	msg_done
	call	sendser
	incf	foo,F
	goto	msg_cont

msg_done:
	bcf	PORTA,SQPIN	; startup is done, shut up

;; the worker loop consumes and interprets the data the interrup handlers
;; have created. all things that take a while are done here.
workerloop:
	; handle pending serial output
	btfss	PIR1,TXIF	; txif is 1 if ready/free
	goto	handle_input
	FEMPTY	txr,txw		; transmission fifo empty?
	btfsc	STATUS,Z	;skips if nonzero
	goto	handle_input
	call	gett
	movwf	TXREG		; safe as we checked for txif ready before

	; handle pending serial input
handle_input:
	FEMPTY	rcvr,rcvw	; serial receive fifo empty?
	btfsc	STATUS,Z	; skips if nonzero
	goto	handle_ir		; nothing pending, check infrared

	bcf	INTCON,GIE	; no ints while modifying the stack
	call	getr		; handle the serial input
	bsf	INTCON,GIE	; ints are fine again
	movwf	baz		; save data in baz for a bit

	; check for existing escapes
	btfsc	cmdstatus,CMDBIT	; this a cmd?
	goto	handle_cmd
	btfsc	cmdstatus,BACKBIT	; backlight on/off?
	goto	handle_backlight
	btfsc	cmdstatus,BEEPBIT	; beeper?
	goto	handle_beeper
	btfsc	cmdstatus,DATABIT	; data?
	goto	isdata

	; no existing escapes, check if this is one
	; data is in baz
	movlw	CMDESC
	subwf	baz,W
	btfsc	STATUS,Z	; equal to cmdesc? skip if !=
	goto	new_cmd
	movlw	BACKESC
	subwf	baz,W
	btfsc	STATUS,Z	; equal to backesc? skip if !=
	goto	new_back
	movlw	BEEPESC
	subwf	baz,W
	btfsc	STATUS,Z	; equal to beepesc? skip if !=
	goto	new_beep
	movlw	DATAESC
	subwf	baz,W
	btfsc	STATUS,Z	; equal to dataesc? skip if !=
	goto	new_data

	; nope, very normal data, so send it to the lcd and be done
isdata:
	clrf	cmdstatus
	movfw	baz
	bcf	STATUS,C	; data!
	call	sendlcd		; that takes a bit over 100us
	goto	workerloop	; done.

; remember received escape codes for the next run
new_beep:
	bsf	cmdstatus,BEEPBIT
	goto	workerloop
new_back:
	bsf	cmdstatus,BACKBIT
	goto	workerloop
new_cmd:
	bsf	cmdstatus,CMDBIT
	goto	workerloop
new_data:
	bsf	cmdstatus,DATABIT
	goto	workerloop

; deal with what follows an escape char
; data is still in baz and w
handle_cmd:
	clrf	cmdstatus
	bsf	STATUS,C
	call	sendlcd		; send the cmd to the lcd
	goto	workerloop
handle_backlight:
	clrf	cmdstatus
	andlw	0xff		; new backlight state: is it zero?
	btfsc	STATUS,Z	; skip if non-zero
	goto	clearback	; zero
	bsf	PORTA,BLPIN	; nonzero, switch it on!
	goto	workerloop
clearback:
	bcf	PORTA,BLPIN
	goto	workerloop
handle_beeper:
	clrf	cmdstatus
	andlw	0xff
	btfsc	STATUS,Z
	goto	clearbeep
	bsf	PORTA,SQPIN
	goto	workerloop
clearbeep:
	bcf	PORTA,SQPIN
	goto	workerloop

handle_ir:
	btfsc	irstatus,IRERR	; was an error reported?
	goto	onerr
	btfsc	irstatus,IRDONE	; valid ir data received?
	goto	save
	btfss	irstatus,IRRPT	; or did we get a repeat?
	goto	workerloop	; neither, so back to work.

	; rpt handling
	bcf	irstatus,IRRPT	; ack the repeat
	; but inc repeat count only iff there's a previous code
	clrw
	iorwf	ircode1,W
	iorwf	ircode2,W
	JNZ	rep		; previous code is ok if any bits are set
	goto	workerloop	; nope, no repeat.
	
	; save the data bytes away, ignoring the first two as fixed 
save:	movfw	d3
	movwf	ircode1
	movfw	d4
	movwf	ircode2
	bcf	irstatus,IRDONE	; ack copy over, d1..4 are free for inthandler
	clrf	ircount

	; and do a bit of sanity checking: ircode1 and 2 are bit-inverses of each other
	movfw	ircode1
	xorwf	ircode2,W	; should be ff
	xorlw	0xff		; complement
	JZ	rep		; ok, incr repeat count and send out data

onerr:	bcf	irstatus,IRERR
	clrf	ircount		; nope - blank it all
	clrf	ircode1
	clrf	ircode2
	goto	workerloop

rep:	incf	ircount,F

	; send the ir codes via serial
	; we do this in hex and one per line for convenience (human-readable stuff rules!)
	; and because it does not really cost too much
	movfw	ircount
	call	sendhex_ser
	movlw	' '
	call	putt
	movfw	ircode1
	call	sendhex_ser
	movfw	ircode2
	call	sendhex_ser
	movlw	0x0d
	call	putt
	movlw	0x0a
	call	putt
	goto	workerloop

; context saving is done. we're in bank0.
; this function returns normally, calling code restores state and retfies
inthandler:
; is it a timer overflow event?
	btfss	INTCON,T0IE	; skips if int is enabled
	goto	test_ir		; tmr0 no int, so certainly not timer-related
	btfss	INTCON,T0IF	; timer overflow? the ir capture has failed, abort it
	goto	test_ir
	bcf	INTCON,T0IE	; disable the tmr0 interrupt	
	bcf	INTCON,T0IF	; ack the interrupt
	bsf	irstatus,IRIDLE ; and abort ir ops
; check rb0 int
test_ir:
	btfss	INTCON,INTE	; rb0 int enabled? skips if on
	goto	test_ser
	btfsc	INTCON,INTF	; rb0 interrupt ? (skip if clear, ie. other)
	call	int_infrared	; yes, handle infrared, then continue here
; check if serial int
test_ser:
	bsf	STATUS,RP0	; bank1
	btfss	PIE1,RCIE	; ser receive int enabled? skip if set
	retlw	0		; lazy, returning in bank1 but context restore will fix it
	bcf	STATUS,RP0	; back in bank0 
	btfss	PIR1,RCIF	; have we received something via serial?
	retlw	0

	; serial reception: first check and clear errors where possible
	btfsc	RCSTA,FERR	; framing error clears on read (but we lose the data)
	goto	is_ferr
	btfss	RCSTA,OERR	; overflow? dis and ena, then read
	goto	is_fine
	; overflow: disable/reenable the receiver, then continue reading
	bcf	RCSTA,CREN
	bsf	RCSTA,CREN
	clrf	cmdstatus	; no sense misinterpreting lost commands

is_fine:	; receive fine, work on it
	movfw	RCREG
	call	putr		; in inthandler, nothing can interrupt us here
	btfss	PIR1,RCIF	; another char a/v? (rcreg is two-deep fifo)
	retlw	0		; no, thanks, done.
	goto	is_fine		; yes, so try again
	
is_ferr:			; cleared on read, next byte will make it good
	movfw	RCREG
	clrf	cmdstatus	; no sense misinterpreting lost commands
	retlw	0

;; note: int_infrared is called and returns!
int_infrared:	
	bcf	INTCON,INTF	; ack the interrupt
 	; figure out if this is the first edge of the code:
	; no int set means we're not counting
	btfsc	INTCON,T0IE	; skip if timer int off
	; yes, then capture, reset tmr, analyze...
	goto	do_cap
	; no, then clear tmr and int on and done
	clrf	TMR0		; blank the timer, as it is running always anyway
	bcf	INTCON,T0IF	; clear overflow flag
	bsf	INTCON,T0IE	; and enable timer0 interrupt
	bsf	irstatus,IRIDLE	; now waiting for HDR or RPT codes
	retlw	0

; do the actual tmr capture and analyze the signal
do_cap:	movfw	TMR0
	clrf	TMR0
	clrf	bar		; need that later

; identify the type of signal, tmr0 in w
; one tmr0 tick is 128us
ZEROLO	EQU	.8		; anything below 1024us is dud
ZEROHI	EQU	.10		; 1024 to to 1280us is a zero
ONELO	EQU	.16
ONEHI	EQU	.19		; 2048 to 2432us is a one
RPTLO	EQU	.86		
RPTHI	EQU	.90		; 11008us to 11520us is repeat
HDRLO	EQU	.103
HDRHI	EQU	.107		; and 13184us to 13696us is header
	; very short: dud
	addlw	-ZEROLO
	btfss	STATUS,C	; skip if non-neg
	goto	irerror
	; is it a zero?
	addlw	-ZEROHI+ZEROLO
	btfss	STATUS,C	; skip if non-neg
	goto	havebit
	; between zero and one?
	addlw	-ONELO+ZEROHI	
	btfss	STATUS,C
	goto	irerror
	; is it a one?
	addlw	-ONEHI+ONELO
	btfss	STATUS,C	; skip if non-neg
	goto	is_one
	; above one but not a repeat? dud
	addlw	-RPTLO+ONEHI	; re-add
	btfss	STATUS,C	; skip if non-neg
	goto	irerror
	; a real repeat?
	addlw	-RPTHI+RPTLO	; and readd again
	btfss	STATUS,C	;skip if non-neg
	goto	is_rpt
	; between repeat and hdr?
	addlw	-HDRLO+RPTHI
	btfss	STATUS,C	;skip if non-neg
	goto	irerror
	; or is it a hdr?
	addlw	-HDRHI+HDRLO
	btfss	STATUS,C
	goto	is_hdr
	; otherwise...it is a dud. let's start from scratch

irerror: 	; something has gone wrong with the ir reception, so we go back to square 1
	bsf	irstatus,IRERR	; notify the worker that repeats make no sense as we've aborted
irdone:	bcf	INTCON,T0IE	; disable the tmr0 interrupt	
	bcf	INTCON,T0IF	; clear tmr0 interrupt flag
	bsf	irstatus,IRIDLE ; and abort ir ops
	retlw	0

; deal with the event order now
is_hdr:	btfss	irstatus,IRIDLE	; got hdr but not idle? dud!
	goto	irerror
	; has the client read any prior result and can we write to d1..d4?
	btfsc	irstatus,IRDONE	; need to skip this code if irdone still set
	goto	irerror
	
	bcf	irstatus,IRIDLE	; we're not idle anymore
	movlw	.32		; 32 bits to go
	movwf	togo
	clrf	d1		; nothing in there now, required for the shifting
	clrf	d2
	clrf	d3
	clrf	d4
	retlw	0		; and we are done

is_rpt:
	btfsc	irstatus,IRIDLE	; got rpt but not idle? dud!
	bsf	irstatus,IRRPT	; signal that we've got a repeat
	goto	irdone		; and switch the timer off, as nothing follows a repeat!
	
is_one:	bsf	bar,0
havebit:
	btfsc	irstatus,IRIDLE	; got 0/1 but idle? dud!
	goto	irerror
	movfw	togo		; got 0/1 but no more required? dud!
	JZ	irerror

	; all good, so lets shift the bit from bar into the right spot
	movlw	d1
	movwf	FSR
	; 32..25 remaining? add to d1
	movlw	.25
	subwf	togo,W
	JNNEG	do_add
	; 24..17? d2
	incf	FSR,F
	movlw	.17
	subwf	togo,W
	JNNEG	do_add
	; 16..9? d3
	incf	FSR,F
	movlw	.9
	subwf	togo,W
	btfss	STATUS,C	; carry set, zero or pos, so stay with d3
	; add it to d4 then.
	incf	FSR,F
	
do_add:	bcf	STATUS,C	
	rlf	INDF,F		; make space, from carry
	movfw	bar		; lowest bit is set or not, others are clear
	iorwf	INDF,F		; and or it in.

	decfsz	togo,F		; one less remaining bits; none left? then we're done!
	retlw	0		
	bsf	irstatus,IRDONE	; data valid, cmd was received
	bcf	irstatus,IRRPT	; it's not a repeat
	bcf	irstatus,IRERR	; and it's a success.
	goto	irdone		; and this resets iridle and disables the timer
	
; this is for sending commands only, during the early init/reset phase.
; input in w, upper half
; porta and b are expected to be off
sendnibble:
	bsf	PORTA,RSPIN	; this seems to help syncing
	bcf	PORTA,RSPIN	; cmd, clear rs 
	bsf	PORTA,ENAPIN	; get the trigger ready

	; ideally i'd do a read-mask-write here, 
	; but too lazy, and with b0..3 not set as outputs it's ok
	andlw	0xf0		; only upper half counts
	movwf	PORTB		; data out
	movlw	7
	call	delayX5us	; this seems critical again, the jdm-204a
				; does not cooperate with less delay
	bcf	PORTA,ENAPIN	; the trigger
	movlw	.10
	call	delayX5us	; and 50us on this side
	clrf	PORTB		; cleanup
	bcf	PORTA,RSPIN
	retlw 0

; input in w, carry bit must be set for command and cleared for data
; needs scritch as tempvar
; destroys w, changes carry
sendlcd:
	bsf	PORTA,RSPIN	; seems to help sync'ing
	btfsc	STATUS,C	; carry is set? then it's a command
	bcf	PORTA,RSPIN	; cmd, hence clear rs
	bsf	PORTA,ENAPIN	; get the trigger ready	

	; ideally i'd do a read-mask-write here, 
	; but too lazy, and with b0..3 not set as outputs it's ok
	movwf	scritch
	andlw	0xf0
	movwf	PORTB	; data out
	bcf	PORTA,ENAPIN	; the trigger

	movlw	.10
	call	delayX5us	; this is critical; my jdm-204a does not
				; work otherwise

	bsf	PORTA,ENAPIN	; second nibble, get the trigger ready	
	swapf	scritch,W
	andlw	0xf0
	movwf	PORTB	; data out
	bcf	PORTA,ENAPIN	; the trigger

	movlw	.10
	call	delayX5us	; 40us is what most cmds need, we do 50us 
	clrf	PORTB
	bcf	PORTA,RSPIN
	retlw	0

; converts byte to two hex chars, puts into transmission buffer
; needs  scratch, foo as tempvars
sendhex_ser:	
	movwf	scratch
	movlw	2
	movwf	foo		; runs
	swapf	scratch,W	; upper nibble
ser_lowernib:
	andlw	0x0f		; only the nibble
	addlw	-.10		; gap in ascii between 0-9 and A-F
	btfsc	STATUS,C	; <10?, skip if neg
	addlw	'A'-'0'-.10	; A-F, but we will do the next op!
	addlw	'0'+.10		; 0-9, but we subtracted 10

	call	putt
	movfw	scratch
	decfsz	foo,F
	goto	ser_lowernib
	retlw	0

; functions to access the tx and rx fifos
getr:	FGETW	rcvr,rcvw
putr:	FPUTW	rcvr,rcvw
	
gett:	FGETW	txr,txw
putt:	FPUTW	txr,txw

	


#include "delay.inc"	

	org	0x2100		; eeprom
; a slight hack: gpasm doesn't produce suitable offset into eeprom area
startmsg	equ $-0x2100	
	de	"ok $Id: dervish.asm,v 2.5 2011/04/25 04:17:49 az Exp $\r\n",0

end

